@@ -1,14 +1,14 @@
 /*
 
 Licensed under the Apache License, Version 2.0 (the "License"); you may not use
-this file except in compliance with the License.  You may obtain a copy of the
+this file except in compliance with the License. You may obtain a copy of the
 License at
 
   http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software distributed
 under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
-CONDITIONS OF ANY KIND, either express or implied.  See the License for the
+CONDITIONS OF ANY KIND, either express or implied. See the License for the
 specific language governing permissions and limitations under the License.
 
 */
@@ -19,6 +19,7 @@ specific language governing permissions and limitations under the License.
 #include "curlhelper.h"
 #include <jsapi.h>
 #include <curl/curl.h>
+#include "config.h"
 
 #ifndef CURLOPT_COPYPOSTFIELDS
    #define CURLOPT_COPYPOSTFIELDS 10165
@@ -214,6 +215,11 @@ EvalInContext(JSContext *context, JSObject *obj, uintN argc, jsval *argv,
         return JS_FALSE;
     }
 
+#ifdef USE_JS_SETOPCB
+    JS_SetContextThread(sub_context);
+    JS_BeginRequest(sub_context);
+#endif
+
     src = JS_GetStringChars(str);
     srclen = JS_GetStringLength(str);
 
@@ -229,11 +235,17 @@ EvalInContext(JSContext *context, JSObject *obj, uintN argc, jsval *argv,
         *rval = OBJECT_TO_JSVAL(sandbox);
         ok = JS_TRUE;
     } else {
-        ok = JS_EvaluateUCScript(sub_context, sandbox, src, srclen, NULL, -1,
+        ok = JS_EvaluateUCScript(sub_context, sandbox, src, srclen, NULL, 0,
                                  rval);
+        ok = JS_TRUE;
     }
 
 out:
+#ifdef USE_JS_SETOPCB
+    JS_EndRequest(sub_context);
+    JS_ClearContextThread(sub_context);
+#endif
+
     JS_DestroyContext(sub_context);
     return ok;
 }
@@ -246,13 +258,13 @@ GC(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
 
 static JSBool
 Print(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
-    uintN i, n;
+    uintN i;
     size_t cl, bl;
     JSString *str;
     jschar *chars;
     char *bytes;
 
-    for (i = n = 0; i < argc; i++) {
+    for (i = 0; i < argc; i++) {
         str = JS_ValueToString(context, argv[i]);
         if (!str)
             return JS_FALSE;
@@ -269,9 +281,8 @@ Print(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval) {
         fprintf(stdout, "%s%s", i ? " " : "", bytes);
         JS_free(context, bytes);
     }
-    n++;
-    if (n)
-        fputc('\n', stdout);
+
+    fputc('\n', stdout);
     fflush(stdout);
     return JS_TRUE;
 }
@@ -395,6 +406,16 @@ ExecuteScript(JSContext *context, JSObject *obj, const char *filename) {
 
 static uint32 gBranchCount = 0;
 
+#ifdef USE_JS_SETOPCB
+static JSBool
+OperationCallback(JSContext *context)
+{
+    if ((++gBranchCount & 0x3fff) == 1) {
+        JS_MaybeGC(context);
+    }
+    return JS_TRUE;
+}
+#else
 static JSBool
 BranchCallback(JSContext *context, JSScript *script) {
     if ((++gBranchCount & 0x3fff) == 1) {
@@ -402,6 +423,7 @@ BranchCallback(JSContext *context, JSScript *script) {
     }
     return JS_TRUE;
 }
+#endif
 
 static void
 PrintError(JSContext *context, const char *message, JSErrorReport *report) {
@@ -434,21 +456,21 @@ typedef struct buffer_counter {
 }* BufferCount;
 
 size_t curl_read(void *ptr, size_t size, size_t nmemb, void *stream) {
-  if( size == 0 || nmemb == 0) {
-    return 0;
-  }
-
+  int readlength, spaceleft, i;
   char* databuffer = (char*)ptr;
   Buffer b = ((BufferCount)stream)->buffer;
   int* pos = &(((BufferCount)stream)->pos);
 
+  if( size == 0 || nmemb == 0) {
+    return 0;
+  }
+
   if((b->count - *pos) == 0) {
     return 0;
   }
 
-  int readlength = size*nmemb;
-  int spaceleft = b->count - *pos;
-  int i;
+  readlength = size*nmemb;
+  spaceleft = b->count - *pos;
 
   if(readlength < spaceleft) {
     copy_Buffer(b,databuffer,*pos,readlength);
@@ -462,11 +484,10 @@ size_t curl_read(void *ptr, size_t size, size_t nmemb, void *stream) {
 }
 
 size_t curl_write(void *ptr, size_t size, size_t nmemb, void *stream) {
-  if( size == 0 || nmemb == 0 )
-    return 0;
-
   char *data, *tmp;
   Buffer b;
+  if( size == 0 || nmemb == 0 )
+    return 0;
 
   data = (char *)ptr;
   b = (Buffer)stream;
@@ -478,23 +499,21 @@ size_t curl_write(void *ptr, size_t size, size_t nmemb, void *stream) {
 
 // This uses MALLOC dont forget to free
 char* JSValToChar(JSContext* context, jsval* arg) {
+  char *c, *tmp;
+  JSString *jsmsg;
+  size_t len;
+  int i;
   if(!JSVAL_IS_STRING(*arg)) {
     return NULL;
   }
 
-  char *c, *tmp;
-  JSString *jsmsg;
-  size_t len;
-
   jsmsg = JS_ValueToString(context,*arg);
   len = JS_GetStringLength(jsmsg);
   tmp = JS_GetStringBytes(jsmsg);
-      
+
   c = (char*)malloc(len+1);
   c[len] = '\0';
 
-  int i;
- 
   for(i = 0;i < len;i++) {
     c[i] = tmp[i];
   }
@@ -534,42 +553,48 @@ struct curl_slist* generateCurlHeaders(JSContext* context,jsval* arg) {
 
     struct curl_slist *slist = NULL;
     JSObject* header_obj;
+    JSObject* iterator;
+    jsval *jsProperty;
+    jsval *jsValue;
+    jsid *jsId;
+    Buffer bTmp;
+    char* jsPropertyName, *jsPropertyValue;
 
     // If we fail to convert arg2 to an object. Error!
     if(!JS_ValueToObject(context,*arg,&header_obj)) {
       return NULL;
     }
 
-    JSObject* iterator = JS_NewPropertyIterator(context,header_obj);
-    
-    jsval *jsProperty = JS_malloc(context,sizeof(jsval));
-    jsval *jsValue = JS_malloc(context,sizeof(jsval));
-    jsid *jsId = JS_malloc(context,sizeof(jsid));
-    
+    iterator = JS_NewPropertyIterator(context,header_obj);
+
+    jsProperty = JS_malloc(context,sizeof(jsval));
+    jsValue = JS_malloc(context,sizeof(jsval));
+    jsId = JS_malloc(context,sizeof(jsid));
+
     while(JS_NextProperty(context,iterator,jsId) == JS_TRUE) {
 
       if(*jsId == JSVAL_VOID) {
-	break;
+        break;
       }
 
       // TODO: Refactor this maybe make a JSValAppendBuffer method b/c that is what you really want to do.
 
-      Buffer bTmp = init_Buffer();
+      bTmp = init_Buffer();
       JS_IdToValue(context,*jsId,jsProperty);
-      char* jsPropertyName = JSValToChar(context,jsProperty);
+      jsPropertyName = JSValToChar(context,jsProperty);
 
       // TODO: Remove strlen =/
       append_Buffer(bTmp,jsPropertyName,strlen(jsPropertyName));
       append_Buffer(bTmp,": ",2);
 
       JS_GetProperty(context,header_obj,jsPropertyName,jsValue);
-      char* jsPropertyValue = JSValToChar(context,jsValue);
+      jsPropertyValue = JSValToChar(context,jsValue);
       // TODO: Remove strlen =/
       append_Buffer(bTmp,jsPropertyValue,strlen(jsPropertyValue));
       append_Buffer(bTmp,"",1);
 
       slist = curl_slist_append(slist,bTmp->data);
-      
+
       free_Buffer(bTmp);
       free(jsPropertyValue);
       free(jsPropertyName);
@@ -592,10 +617,12 @@ GetHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   Buffer b;
   char *url;
   size_t charslen, readlen;
+  struct curl_slist *slist;
+  int exitcode;
 
   // Run GC
   JS_MaybeGC(context);
-  
+
   // Init Curl
   if((handle = curl_easy_init()) == NULL) {
     return JS_FALSE;
@@ -616,18 +643,16 @@ GetHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   curl_easy_setopt(handle,CURLOPT_WRITEHEADER,b);
   curl_easy_setopt(handle,CURLOPT_URL,url);
   curl_easy_setopt(handle,CURLOPT_HTTPGET,1);
-  curl_easy_setopt(handle,CURLOPT_FOLLOWLOCATION,1);  
+  curl_easy_setopt(handle,CURLOPT_FOLLOWLOCATION,1);
   curl_easy_setopt(handle,CURLOPT_NOPROGRESS,1);
   curl_easy_setopt(handle,CURLOPT_IPRESOLVE,CURL_IPRESOLVE_V4);
 
-  struct curl_slist *slist = generateCurlHeaders(context,argv+1);
+  slist = generateCurlHeaders(context,argv+1);
   if(slist != NULL) {
     curl_easy_setopt(handle,CURLOPT_HTTPHEADER,slist);
   }
 
   // Perform
-  int exitcode;
-
   if((exitcode = curl_easy_perform(handle)) != 0) {
     if(slist != NULL) {
       curl_slist_free_all(slist);
@@ -654,7 +679,7 @@ GetHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   /* Shrink the buffer to the real size  and store its value in rval */
   shrink_Buffer(b);
   BufferToJSVal(context,b,rval);
-  
+
   // Free Buffer
   free_Buffer(b);
 
@@ -676,10 +701,12 @@ HeadHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   Buffer b;
   char *url;
   size_t charslen, readlen;
+  struct curl_slist *slist;
+  int exitcode;
 
   // Run GC
   JS_MaybeGC(context);
-  
+
   // Init Curl
   if((handle = curl_easy_init()) == NULL) {
     return JS_FALSE;
@@ -704,7 +731,7 @@ HeadHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   curl_easy_setopt(handle,CURLOPT_NOPROGRESS,1);
   curl_easy_setopt(handle,CURLOPT_IPRESOLVE,CURL_IPRESOLVE_V4);
 
-  struct curl_slist *slist = generateCurlHeaders(context,argv+1);
+  slist = generateCurlHeaders(context,argv+1);
   if(slist != NULL) {
     curl_easy_setopt(handle,CURLOPT_HTTPHEADER,slist);
   }
@@ -712,8 +739,6 @@ HeadHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   // fprintf(stderr, "about to run HEAD request\n");
 
   // Perform
-  int exitcode;
-
   if((exitcode = curl_easy_perform(handle)) != 0) {
     if(slist != NULL) {
       curl_slist_free_all(slist);
@@ -741,7 +766,7 @@ HeadHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   /* Shrink the buffer to the real size  and store its value in rval */
   shrink_Buffer(b);
   BufferToJSVal(context,b,rval);
-  
+
   // Free Buffer
   free_Buffer(b);
 
@@ -764,6 +789,8 @@ PostHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   Buffer b;
   char *url, *body;
   size_t charslen, readlen;
+  struct curl_slist *slist;
+  int exitcode;
 
   // Run GC
   JS_MaybeGC(context);
@@ -801,13 +828,11 @@ PostHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   curl_easy_setopt(handle,CURLOPT_POSTFIELDSIZE,strlen(body));
   curl_easy_setopt(handle,CURLOPT_POSTFIELDS,body);         // Curl wants '\0' terminated, we oblige
 
-  struct curl_slist *slist = generateCurlHeaders(context,argv+2); // Initialize Headers
+  slist = generateCurlHeaders(context,argv+2); // Initialize Headers
   if(slist != NULL) {
-    curl_easy_setopt(handle,CURLOPT_HTTPHEADER,slist);            
+    curl_easy_setopt(handle,CURLOPT_HTTPHEADER,slist);
   }
 
-  int exitcode;
-
   if((exitcode = curl_easy_perform(handle)) != 0) {             // Perform
     curl_slist_free_all(slist);
     free(body);
@@ -838,7 +863,7 @@ PostHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
 #define CLEAN \
   free_Buffer(b); \
   free_Buffer(b_data->buffer); \
-  free(b_data);		       \
+  free(b_data);                \
   free(url)
 
 static JSBool
@@ -849,6 +874,9 @@ PutHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   char *url, *data;
   size_t charslen, readlen;
   JSObject* header_obj;
+  CURL* handle;
+  struct curl_slist *slist;
+  int exitcode;
 
   // Run GC
   JS_MaybeGC(context);
@@ -858,24 +886,22 @@ PutHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
 
   // Allocate buffer that will store the get resultant
   b = init_Buffer();
-  
+
   // Allocate data buffer and move data into them
   b_data = (BufferCount)malloc(sizeof(Buffer) + sizeof(int));
   b_data->buffer = init_Buffer();
   b_data->pos = 0;
-  
+
   data = JSValToChar(context,(argv+1));
   readlen = strlen(data);
-  
 
-  
+
+
   // TODO: remove strlen
   append_Buffer(b_data->buffer,data,readlen);
 
   free(data);
 
-  CURL* handle;
-
   // Init Curl
 
   if((handle = curl_easy_init()) == NULL) {
@@ -893,11 +919,11 @@ PutHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   curl_easy_setopt(handle,CURLOPT_URL,url);
   curl_easy_setopt(handle,CURLOPT_UPLOAD,1);
   curl_easy_setopt(handle,CURLOPT_INFILESIZE,readlen);
-  
+
 
 
   // Curl structure
-  struct curl_slist *slist = generateCurlHeaders(context,argv+2);
+  slist = generateCurlHeaders(context,argv+2);
   if(slist != NULL) {
     curl_easy_setopt(handle,CURLOPT_HTTPHEADER,slist);
   }
@@ -908,11 +934,9 @@ PutHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   // Use only ipv4
   curl_easy_setopt(handle,CURLOPT_IPRESOLVE,CURL_IPRESOLVE_V4);
 
-  
+
 
   // Perform
-  int exitcode;
-  
   if((exitcode = curl_easy_perform(handle)) != 0) {
     if(slist != NULL)
       curl_slist_free_all(slist);
@@ -939,7 +963,7 @@ PutHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   shrink_Buffer(b);
 
   BufferToJSVal(context,b,rval);
-  
+
   free_Buffer(b);
 
   if(rval == NULL) {
@@ -960,6 +984,10 @@ DelHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   char *url;
   size_t charslen, readlen;
   char header_name[7];
+  CURL* handle;
+  int exitcode;
+  struct curl_slist *slist = NULL;
+
   strcpy(header_name,"DELETE");
 
   // Run GC
@@ -971,8 +999,6 @@ DelHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   // Allocate buffer that will store the del resultant
   b = init_Buffer();
 
-  CURL* handle;
-
   // Init Curl
   if((handle = curl_easy_init()) == NULL) {
     free_Buffer(b);
@@ -990,14 +1016,11 @@ DelHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   curl_easy_setopt(handle,CURLOPT_IPRESOLVE,CURL_IPRESOLVE_V4);
 
   // Curl structure
-  struct curl_slist *slist = NULL;
   if((slist = generateCurlHeaders(context,argv+1)) != NULL) {
     curl_easy_setopt(handle,CURLOPT_HTTPHEADER,slist);
   }
 
   // Perform
-  int exitcode;
-
   if((exitcode = curl_easy_perform(handle)) != 0) {
     if(slist != NULL)
       curl_slist_free_all(slist);
@@ -1023,7 +1046,7 @@ DelHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
   shrink_Buffer(b);
 
   BufferToJSVal(context,b,rval);
-  
+
   if(rval == NULL) {
     curl_easy_cleanup(handle);
     return JS_FALSE;
@@ -1042,6 +1065,10 @@ CopyHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   char *url;
   size_t charslen, readlen;
   char header_name[5];
+  CURL* handle;
+  int exitcode;
+  struct curl_slist *slist = NULL;
+
   strcpy(header_name,"COPY");
 
   // Run GC
@@ -1053,8 +1080,6 @@ CopyHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   // Allocate buffer that will store the del resultant
   b = init_Buffer();
 
-  CURL* handle;
-
   // Init Curl
   if((handle = curl_easy_init()) == NULL) {
     free_Buffer(b);
@@ -1072,14 +1097,11 @@ CopyHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   curl_easy_setopt(handle,CURLOPT_IPRESOLVE,CURL_IPRESOLVE_V4);
 
   // Curl structure
-  struct curl_slist *slist = NULL;
   if((slist = generateCurlHeaders(context,argv+1)) != NULL) {
     curl_easy_setopt(handle,CURLOPT_HTTPHEADER,slist);
   }
 
   // Perform
-  int exitcode;
-
   if((exitcode = curl_easy_perform(handle)) != 0) {
     if(slist != NULL)
       curl_slist_free_all(slist);
@@ -1105,7 +1127,7 @@ CopyHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   shrink_Buffer(b);
 
   BufferToJSVal(context,b,rval);
-  
+
   if(rval == NULL) {
     curl_easy_cleanup(handle);
     return JS_FALSE;
@@ -1124,6 +1146,10 @@ MoveHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   char *url;
   size_t charslen, readlen;
   char header_name[5];
+  CURL* handle;
+  struct curl_slist *slist = NULL;
+  int exitcode;
+
   strcpy(header_name,"MOVE");
 
   // Run GC
@@ -1135,8 +1161,6 @@ MoveHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   // Allocate buffer that will store the del resultant
   b = init_Buffer();
 
-  CURL* handle;
-
   // Init Curl
   if((handle = curl_easy_init()) == NULL) {
     free_Buffer(b);
@@ -1154,14 +1178,11 @@ MoveHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   curl_easy_setopt(handle,CURLOPT_IPRESOLVE,CURL_IPRESOLVE_V4);
 
   // Curl structure
-  struct curl_slist *slist = NULL;
   if((slist = generateCurlHeaders(context,argv+1)) != NULL) {
     curl_easy_setopt(handle,CURLOPT_HTTPHEADER,slist);
   }
 
   // Perform
-  int exitcode;
-
   if((exitcode = curl_easy_perform(handle)) != 0) {
     if(slist != NULL)
       curl_slist_free_all(slist);
@@ -1187,7 +1208,7 @@ MoveHttp(JSContext *context, JSObject *obj, uintN argc, jsval *argv, jsval *rval
   shrink_Buffer(b);
 
   BufferToJSVal(context,b,rval);
-  
+
   if(rval == NULL) {
     curl_easy_cleanup(handle);
     return JS_FALSE;
@@ -1212,9 +1233,16 @@ main(int argc, const char * argv[]) {
     context = JS_NewContext(runtime, gStackChunkSize);
     if (!context)
         return 1;
+    /* FIXME: https://bugzilla.mozilla.org/show_bug.cgi?id=477187 */
     JS_SetErrorReporter(context, PrintError);
+#ifdef USE_JS_SETOPCB
+    JS_SetContextThread(context);
+    JS_BeginRequest(context);
+    JS_SetOperationCallback(context, OperationCallback);
+#else
     JS_SetBranchCallback(context, BranchCallback);
     JS_ToggleOptions(context, JSOPTION_NATIVE_BRANCH_CALLBACK);
+#endif
     JS_ToggleOptions(context, JSOPTION_XML);
 
     global = JS_NewObject(context, NULL, NULL, NULL);
@@ -1245,6 +1273,11 @@ main(int argc, const char * argv[]) {
 
     ExecuteScript(context, global, argv[1]);
 
+#ifdef USE_JS_SETOPCB
+    JS_EndRequest(context);
+    JS_ClearContextThread(context);
+#endif
+
     JS_DestroyContext(context);
     JS_DestroyRuntime(runtime);
     JS_ShutDown();
